module Loadable exposing
    ( Loadable(..), expectUpdate, update
    , isLoading, isStale, hasValue, hasError
    , map, map2, map3, mapError, withDefault, toValue, toError, fromResult
    )

{-| This module helps you model values loaded, for example, from a database over Http/GraphQL.

@docs Loadable, expectUpdate, update


# Queries

@docs isLoading, isStale, hasValue, hasError


# Conversion

@docs map, map2, map3, mapError, withDefault, toValue, toError, fromResult

-}


{-| A representation of data-loading. This would usually go into your model.
-}
type Loadable error value
    = Idle
    | Loading
    | Success value
    | Failure error
    | Reloading value
    | ReloadFailure error value


{-| Update your loadable to expect a new value. Use this when you expect a new value to arrive soon.
-}
expectUpdate : Loadable error value -> Loadable error value
expectUpdate loadable =
    case loadable of
        Success value ->
            Reloading value

        ReloadFailure _ stale ->
            Reloading stale

        Failure _ ->
            Loading

        Idle ->
            Loading

        Loading ->
            loadable

        Reloading _ ->
            loadable


{-| Update the loadable with a result. Use this when a new value arrives, and your loadable is expecting the update.
-}
update : Result error value -> Loadable error value -> Loadable error value
update result loadable =
    case result of
        Ok value ->
            Success value

        Err error ->
            case loadable of
                Success value ->
                    ReloadFailure error value

                ReloadFailure _ stale ->
                    ReloadFailure error stale

                Reloading value ->
                    ReloadFailure error value

                Failure _ ->
                    Failure error

                Idle ->
                    Failure error

                Loading ->
                    Failure error



-- Queries


{-| Is the loadable expecting a new value?
-}
isLoading : Loadable error value -> Bool
isLoading loadable =
    case loadable of
        Loading ->
            True

        Reloading _ ->
            True

        Success _ ->
            False

        ReloadFailure _ _ ->
            False

        Failure _ ->
            False

        Idle ->
            False


{-| Does the loadable contain a stale value?
Staleness implies a reload is underway or failed, in which case the value from the most recent load is available.
-}
isStale : Loadable error value -> Bool
isStale loadable =
    case loadable of
        Reloading _ ->
            True

        ReloadFailure _ _ ->
            True

        Loading ->
            False

        Success _ ->
            False

        Failure _ ->
            False

        Idle ->
            False


{-| Is any value, stale or otherwise, contained within the loadable?
-}
hasValue : Loadable error value -> Bool
hasValue =
    (/=) Nothing << toValue


{-| Has the loadable failed on the latest load or reload?
-}
hasError : Loadable error value -> Bool
hasError =
    (/=) Nothing << toError



-- Conversion


{-| Transform the value of a loadable.
-}
map : (value -> otherValue) -> Loadable error value -> Loadable error otherValue
map fromValue loadable =
    case loadable of
        Idle ->
            Idle

        Loading ->
            Loading

        Success value ->
            Success (fromValue value)

        Failure error ->
            Failure error

        Reloading stale ->
            Reloading (fromValue stale)

        ReloadFailure error stale ->
            ReloadFailure error (fromValue stale)


{-| Transform the values of two loadables, falling back to the weakest constructor of the two. Precedence is as follows:

1.  Failure
2.  Idle
3.  Loading
4.  ReloadFailure
5.  Reloading
6.  Success

-}
map2 :
    (first -> second -> third)
    -> Loadable error first
    -> Loadable error second
    -> Loadable error third
map2 combine first second =
    case first of
        Idle ->
            case second of
                Idle ->
                    Idle

                Loading ->
                    Idle

                Success _ ->
                    Idle

                Failure error ->
                    Failure error

                Reloading _ ->
                    Idle

                ReloadFailure error _ ->
                    Failure error

        Loading ->
            case second of
                Idle ->
                    Idle

                Loading ->
                    Idle

                Success _ ->
                    Loading

                Failure error ->
                    Failure error

                Reloading _ ->
                    Loading

                ReloadFailure _ _ ->
                    Loading

        Success value ->
            case second of
                Idle ->
                    Idle

                Loading ->
                    Loading

                Success secondValue ->
                    Success (combine value secondValue)

                Failure error ->
                    Failure error

                Reloading secondValue ->
                    Reloading (combine value secondValue)

                ReloadFailure error secondValue ->
                    ReloadFailure error (combine value secondValue)

        Failure error ->
            Failure error

        Reloading value ->
            case second of
                Idle ->
                    Idle

                Loading ->
                    Loading

                Success secondValue ->
                    Success (combine value secondValue)

                Failure error ->
                    Failure error

                Reloading secondValue ->
                    Reloading (combine value secondValue)

                ReloadFailure error secondValue ->
                    ReloadFailure error (combine value secondValue)

        ReloadFailure error value ->
            case second of
                Idle ->
                    Idle

                Loading ->
                    Loading

                Success secondValue ->
                    ReloadFailure error (combine value secondValue)

                Failure otherError ->
                    Failure otherError

                Reloading secondValue ->
                    ReloadFailure error (combine value secondValue)

                ReloadFailure _ secondValue ->
                    ReloadFailure error (combine value secondValue)


{-| Refer to [map2](#map2). Loadables are combined chronologically.
-}
map3 :
    (value -> secondValue -> thirdValue -> fourthValue)
    -> Loadable error value
    -> Loadable error secondValue
    -> Loadable error thirdValue
    -> Loadable error fourthValue
map3 combine first second =
    map2 (\( f, s ) t -> combine f s t)
        (map2 Tuple.pair first second)


{-| Transform the error of a loadable.
-}
mapError : (error -> otherError) -> Loadable error value -> Loadable otherError value
mapError fromError loadable =
    case loadable of
        Idle ->
            Idle

        Loading ->
            Loading

        Success value ->
            Success value

        Failure error ->
            Failure (fromError error)

        Reloading stale ->
            Reloading stale

        ReloadFailure error stale ->
            ReloadFailure (fromError error) stale


{-| Get the value of a loadable, or a default value if there is none.
-}
withDefault : value -> Loadable error value -> value
withDefault default loadable =
    case loadable of
        Success value ->
            value

        Idle ->
            default

        Loading ->
            default

        Failure _ ->
            default

        Reloading stale ->
            stale

        ReloadFailure _ stale ->
            stale


{-| Extract the value of a loadable, should there be one.
-}
toValue : Loadable error value -> Maybe value
toValue loadable =
    case loadable of
        Success value ->
            Just value

        ReloadFailure _ stale ->
            Just stale

        Reloading stale ->
            Just stale

        _ ->
            Nothing


{-| Extract the error of a loadable, should there be one.
-}
toError : Loadable error value -> Maybe error
toError loadable =
    case loadable of
        Failure error ->
            Just error

        ReloadFailure error _ ->
            Just error

        _ ->
            Nothing


{-| Transform a result into a loadable.
-}
fromResult : Result error value -> Loadable error value
fromResult result =
    case result of
        Ok value ->
            Success value

        Err error ->
            Failure error
